//#########################################################################
// spifast.s
//
// Write/read a buffer with SPI modul to from MMC/SD/SDHC.
// Optimized fast operation. Only SPI0 modul supported at this time.
//
// New version for ATMega644 which does not support "sbis SPSR, SPIF".
// Uses "in" and "sbrs" instead.
//
//#########################################################################
// Last change: 24.05.2008
//#########################################################################
// hk@holger-klabunde.de
// http://www.holger-klabunde.de/index.html
//#########################################################################

#define __ASSEMBLER__ 1
#define __SFR_OFFSET 0

#include <avr/io.h>

#include "mmc_spi.h"

#ifdef FAST_SPI_WRITE_ASM

/** \ingroup mmc_spi
    \fn void write_spi_fast(size_t len,const void *src)
    \brief Copy a memory area to SPI using a fast method.
           
    \returns Nothing

__tmp_reg__ 	= r0
len		= r24,r25
*src		= r22,r23
      */

	.text
	.global	write_spi_fast
	.type	write_spi_fast, @function
write_spi_fast:
        push r30
        push r31
        
	mov	r30, r22                ; copy adress of buffer to Z
	mov	r31, r23

write_fast_loop:
	ld	r22, Z+                 ;2 get data from memory

wl4:
        nop                             ;1 this speeds up a little bit if F_CPU/2
	in r0,SPSR			;1 on ATMega644 sbis does not work !
	sbrs r0,SPIF                    ;1
       	rjmp	wl4                     ;2 if not skipped
 
	out	SPDR, r22               ;1 write data 

	sbiw	r24, 1          	;2 decrement len
	brne write_fast_loop            ;2 if branch, 1 if not

        ldi	r23, 0xFF		; Dummy byte for spi write
        				; Does not work if dummy byte is not 0xFF !?
wl1:
	in r0,SPSR			;1
	sbrs r0,SPIF                    ;1
       	rjmp	wl1                     ;2 if not skipped

	out	SPDR, r23               ;1 read and ignore CRC16 part1
wl2:
	in r0,SPSR			;1
	sbrs r0,SPIF                    ;1
       	rjmp	wl2                     ;2 if not skipped

	out	SPDR, r23               ;1 read and ignore CRC16 part2
        pop r31                         ;2
        pop r30                         ;2

wl3:
	in r0,SPSR			;1
	sbrs r0,SPIF                    ;1
       	rjmp	wl3                     ;2 if not skipped

	ret                             ;5 !
#endif //FAST_SPI_WRITE_ASM

#ifdef FAST_SPI_READ_ASM
/** \ingroup mmc_spi
    \fn void read_spi_fast(size_t len,const void *dest)
    \brief Copy from SPI to a memory area using a fast method.
           
    \returns Nothing

__tmp_reg__ 	= r0
len		= r24,r25
*dest		= r22,r23
      */
    
	.text
	.global	read_spi_fast
	.type	read_spi_fast, @function
read_spi_fast:

        push r30
        push r31

	mov	r30, r22                ; copy adress of buffer to Z
	mov	r31, r23

        ldi	r23, 0xFF		; Dummy byte for spi write
        				; does not work if dummy byte is not 0xFF !?
	out	SPDR, r23               ; start shift in first byte 
rl1:
	in r0,SPSR			; on ATMega644 sbis does not work !
	sbrs r0,SPIF
       	rjmp	rl1

      	in	r22, SPDR 		; get first byte, but store later
	out	SPDR, r23               ; start shift in next byte
	                                 
	sbiw	r24, 1          	;2 decrement len for first byte read

read_fast_loop:
       	st	Z+, r22			;2 store byte to buffer

rl4:
        nop                             ;1 this speeds up a little bit if F_CPU/2
	in r0,SPSR			;1
	sbrs r0,SPIF                    ;1
       	rjmp	rl4                     ;2

       	in	r22, SPDR               ;1 read last received byte
      	out	SPDR, r23               ;1 start shift in next byte

	sbiw	r24, 1          	;2 decrement len
	brne read_fast_loop             ;2 if branch, 1 if not

       	st	Z, r22 			;2 store last received byte

; last SPDR write is shifting in crc part1 at this point
rl2:
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	rl2

; shift in crc part2
       	out	SPDR, r23
        pop r31                         ; 2
        pop r30                         ; 2

rl3:
	in r0,SPSR			
	sbrs r0,SPIF
       	rjmp	rl3

	ret                             ; 5
#endif //FAST_SPI_READ_ASM

